module PS2_A2Z (
	input clk_input,
	input PS2_clk,
	input PS2_data,
	output reg [7:0] data_out,
	output reg data_ready
); 

	reg PS2_clk_pre;
	reg PS2_clk_buf;
	reg PS2_dat_buf;
	reg [10:0] shift_reg_PS;
	reg [12:0] tps_clk_PS;//13:0
	reg [3:0] count_bit;
	reg [3:0] state_PS2;
	reg timer_OK;
	reg frame_full;
	reg frame_begin_end_ok;
	
	
	always @(*) begin
		if  ( (tps_clk_PS[12:7] >= 6'd10) && (tps_clk_PS[12:7] <= 6'd23) )//13:8
			timer_OK <= 1;
		else 
			timer_OK <= 0;
	end
	
	always @(*) begin
		if (count_bit == 4'd10) 
			frame_full <= 1'b1;
		else 
			frame_full <= 1'b0;
	end
	
	always @(*) begin
		if ( (shift_reg_PS[0] == 1'b0) && (shift_reg_PS[10] == 1'b1) ) 
			frame_begin_end_ok <= 1'b1;
		else 
			frame_begin_end_ok <= 1'b0;
	end
	
	always @(posedge clk_input) begin
		PS2_dat_buf <= PS2_data;
		PS2_clk_buf <= PS2_clk;
		PS2_clk_pre <= PS2_clk_buf;
		
		// STATE 0 reset counter
		if (state_PS2 == 3'd0) begin // Reset counter
			tps_clk_PS <= 13'd0;
			state_PS2 <= 3'd1;
			data_ready <= 1'b0;
		end
		else 
			tps_clk_PS <= tps_clk_PS + 13'd1;
		
		// STATE 1 : waiting for falling edge
		if ( (state_PS2 == 3'd1) && (PS2_clk_buf==1'b0) && (PS2_clk_pre==1'b1) )
			state_PS2 <= 3'd2;
			
		// STATE 2 : shift reg management
		if (state_PS2 == 3'd2) begin
			shift_reg_PS <= {PS2_dat_buf, shift_reg_PS[10:1]};
			state_PS2 <= 3'd3;
		end
		
		// STATE 3 : check timer & increment bit count
		if ( (state_PS2 == 3'd3) && (timer_OK == 1'b0) ) begin
			count_bit <= 4'd0;
			state_PS2 <= 3'd0;
		end
		if ( (state_PS2 == 3'd3) && (timer_OK == 1'b1) ) begin
			count_bit <= count_bit + 4'd1; 
			state_PS2 <= 3'd4;
		end
		
		// STATE 4 : check bit count
		if ( (state_PS2 == 3'd4) && (frame_full == 1'b0) ) 
			state_PS2 <= 3'd0;
		if ( (state_PS2 == 3'd4) && (frame_full == 1'b1) ) 
			state_PS2 <= 3'd5;
			
		// STATE 5 : frame full check begin and end of frame
		if (state_PS2 == 3'd5)
			count_bit <= 4'd15;
		if ( (state_PS2 == 3'd5) && (frame_begin_end_ok == 1'b1) ) begin
			data_out <= shift_reg_PS[8:1];
			//count_bit <= 4'd15;
			state_PS2 <= 3'd6;
		end
		if ( (state_PS2 == 3'd5) && (frame_begin_end_ok == 1'b0) ) begin
			//count_bit <= 4'd15;
			state_PS2 <= 3'd0;
		end
		
		// STATE 6 : pulse "data ready"
		if (state_PS2 == 3'd6) begin
			data_ready <= 1'b1;
			state_PS2 <= 3'd0;
		end
	end
	
endmodule