module essai_VGA (
	input clk_slow, 
	input reset_in, 
	input [7:0] data_from_cpu,
	input [11:0] wr_adress,
	input wr_en,
	input [2:0] micro_counter,
	input [63:0] VGA_buffer, 
	input frame_requested, 
	output reg frame_displayed,
	output [16:0] VGA_read_addr,
	output H_sync_f, 
	output reg V_sync, 
	output [3:0] red, 
	output [3:0] blue, 
	output [3:0] green
) ; 

	reg [10:0] counter_H;
	reg [6:0] counter_Hread;
	reg [6:0] counter_Htext;
	reg [9:0] counter_V;
	reg [11:0] counter_V2;
	reg [2:0] horiz_char;
	reg H_sync;
	reg activ_char_l, activ_char_buf, activ_ligne;
	reg [2:0] pixel_micro_counter;
	
	wire [14:0] adress_charrom;
	wire activ_char, pixel_l;
	wire pixel_char;
	wire [7:0] char_ID;
	wire [11:0] char_adress;
	
	wire[3:0] graph_red;
	wire[3:0] graph_green;
	wire[3:0] graph_blue;
	reg[7:0] cur_pixel;
	
		
	char_ROM char_ROM_l (adress_charrom, clk_slow, pixel_l);

	text_RAM text_RAM1 (data_from_cpu, char_adress, clk_slow, wr_adress, clk_slow, wr_en, char_ID);

	add_VGA_full add_VGA_full_1 (counter_Hread, counter_V2, VGA_read_addr);
	
	VGA_palette palette_red (cur_pixel[7:6], cur_pixel[1:0], graph_red);
	VGA_palette palette_green (cur_pixel[5:4], cur_pixel[1:0], graph_green);
	VGA_palette palette_blue (cur_pixel[3:2], cur_pixel[1:0], graph_blue);
	
	assign adress_charrom={char_ID[7:0],counter_V[3:0],horiz_char[2:0]};
	assign char_adress={counter_V[8:4],counter_Htext};
	assign pixel_char = pixel_l;
	assign H_sync_f=H_sync;
	assign activ_char=activ_ligne && activ_char_buf;
	
	assign red[0]=!(activ_char && (pixel_char || graph_red[0] ));
	assign red[1]=!(activ_char && (pixel_char||graph_red[1] ));
	assign red[2]=!(activ_char && (pixel_char||graph_red[2] ));
	assign red[3]=!(activ_char && (pixel_char||graph_red[3] ));
	assign blue[0]=!(activ_char && (pixel_char||graph_blue[0] ));
	assign blue[1]=!(activ_char && (pixel_char||graph_blue[1] ));
	assign blue[2]=!(activ_char && (pixel_char||graph_blue[2] ));
	assign blue[3]=!(activ_char && (pixel_char||graph_blue[3] ));
	assign green[0]=!(activ_char && (pixel_char||graph_green[0] ));
	assign green[1]=!(activ_char && (pixel_char||graph_green[1] ));
	assign green[2]=!(activ_char && (pixel_char||graph_green[2] ));
	assign green[3]=!(activ_char && (pixel_char||graph_green[3] ));
	
	
	always @(*) begin
		case (micro_counter)
			3'd0 : cur_pixel <= VGA_buffer[39:32];//55:48
			3'd1 : cur_pixel <= VGA_buffer[47:40];//63:56
			3'd2 : cur_pixel <= VGA_buffer[55:48];//7:0
			3'd3 : cur_pixel <= VGA_buffer[63:56];//15:8
			3'd4 : cur_pixel <= VGA_buffer[7:0];//23:16
			3'd5 : cur_pixel <= VGA_buffer[15:8];//31:24
			3'd6 : cur_pixel <= VGA_buffer[23:16];//39:32
			3'd7 : cur_pixel <= VGA_buffer[31:24];//47:40
			default : cur_pixel <= 8'd0;
		endcase
	end
	
	always @(*) begin
		case (micro_counter)
			3'd0 : horiz_char <= 3'd5;
			3'd1 : horiz_char <= 3'd6;
			3'd2 : horiz_char <= 3'd7;
			3'd3 : horiz_char <= 3'd0;
			3'd4 : horiz_char <= 3'd1;
			3'd5 : horiz_char <= 3'd2;
			3'd6 : horiz_char <= 3'd3;
			3'd7 : horiz_char <= 3'd4;
			default : horiz_char <= 3'd6;
		endcase
	end	
	
	always @(posedge clk_slow) begin
	
		if (micro_counter == 3'd1) 
			counter_Htext <= counter_Hread;
	
		if (micro_counter == 3'd3)
			activ_char_buf <= activ_char_l;
	
		if (micro_counter == 3'd5) begin 
			counter_Hread <= counter_Hread+7'd1;
			
			if (counter_Hread==7'd100) // 800/8
				begin
					counter_Hread<=7'd0;
					activ_char_l<=1;
					counter_V<=counter_V+10'd1;
					counter_V2<=counter_V2+12'd3;
				end
			
			if (counter_Hread==7'd82) // 656/8
				H_sync<=0;
			
			if (counter_Hread==7'd94) // 752/8 
				H_sync<=1;
			
			if (counter_Hread==7'd80)	//640/8
				activ_char_l<=0;
		end
		
		if (micro_counter==4'd6) begin
			if (counter_V==10'd525) begin 
				counter_V<=10'd0;
				if (frame_requested==1'b0) begin //Frame buffer 0
					counter_V2<=12'd0;
					frame_displayed <= 1'b0;
				end
				if (frame_requested==1'b1) begin //Frame buffer 1
					counter_V2<=12'h5A0;
					frame_displayed <= 1'b1;
				end
				activ_ligne<=1'b1;
			end
			if (counter_V==10'd480)
				activ_ligne<=1'b0;
			if (counter_V==10'd490)
				V_sync<=1'b0;
			if (counter_V==10'd492)
				V_sync<=1'b1;
		end
	end
	
endmodule