module ALU_A2Z (
	input [15:0] ALU_A,
	input [15:0] ALU_B,
	input [3:0] ALU_OP, 
	output reg [15:0] ALU_result, 
	output reg ALU_cond,
	output reg activ_cond
);

	wire [15:0] addition;
	wire [5:0] comp_res;
	wire[31:0] mult_result;
	wire [15:0] quotient_div;
	wire [15:0] remain_div;
	
	ADD_1 ADD_1(ALU_OP[0],ALU_A, ALU_B, addition);
	
	mult_01 mult_01(ALU_A, ALU_B, mult_result);
	
	division_01 division_01 (ALU_B, ALU_A, quotient_div, remain_div);
	
	compare_01 compare_01(ALU_A, ALU_B, comp_res[0], comp_res[1], comp_res[2], comp_res[3], 
		comp_res[4], comp_res[5]);
	
	
	always @(*) begin			// MUX ALU_result
		case (ALU_OP[3:0])
			4'b0000 : ALU_result <= addition;
			4'b0001 : ALU_result <= addition;
			4'b0010 : ALU_result <= 0; // NOT IMPLEMENTED YET! 
			4'b0011 : ALU_result <= mult_result[31:16];
			4'b0100 : ALU_result <= mult_result[15:0];
			4'b0101 : ALU_result <= quotient_div; 
			4'b0110 : ALU_result <= remain_div; 
			4'b0111 : ALU_result <= ~ALU_A; 
			
			4'b1000 : ALU_result <= {15'd0,ALU_cond};
			4'b1001 : ALU_result <= {15'd0,ALU_cond};
			4'b1010 : ALU_result <= {15'd0,ALU_cond};
			4'b1011 : ALU_result <= {15'd0,ALU_cond};
			4'b1100 : ALU_result <= {15'd0,ALU_cond};
			4'b1101 : ALU_result <= {15'd0,ALU_cond};
			4'b1110 : ALU_result <= ALU_A & ALU_B;
			4'b1111 : ALU_result <= ALU_A | ALU_B;
			default : ALU_result <= 0;
		endcase
	end
	
	always@(*) begin
		case (ALU_OP[3:0])
			4'b0000 : activ_cond <= 1'b0;
			4'b0001 : activ_cond <= 1'b0;
			4'b0010 : activ_cond <= 1'b0;
			4'b0011 : activ_cond <= 1'b0;
			4'b0100 : activ_cond <= 1'b0;
			4'b0101 : activ_cond <= 1'b0;
			4'b0110 : activ_cond <= 1'b0;
			default : activ_cond <= 1'b1;
		endcase
	end
	
	always @(*) begin			// MUX ALU_cond
		case (ALU_OP[3:0])
			4'b0111 : ALU_cond <= ~ALU_A[0];
			4'b1000 : ALU_cond <= comp_res[0]; // aeb		(A==B)
			4'b1001 : ALU_cond <= comp_res[1]; // agb 	(A> B) 
			4'b1010 : ALU_cond <= comp_res[2]; // ageb	(A>=B)
			4'b1011 : ALU_cond <= comp_res[3]; // alb		(A< B)
			4'b1100 : ALU_cond <= comp_res[4]; // aleb	(A<=B)
			4'b1101 : ALU_cond <= comp_res[5]; // aneb 	(A!=B)
			4'b1110 : ALU_cond <= ALU_A[0] & ALU_B[0];
			4'b1111 : ALU_cond <= ALU_A[0] | ALU_B[0];
			default : ALU_cond <= 0;
		endcase
	end

endmodule