module core_A2Z (
	input clk_slow,
	input clk_fast,
	input reset_in,
	input [16:0] addr_VGA_high,
	input [15:0] ext_to_CPU,
	output reg [15:0] X_from_CPU,
	output reg [23:0] addr_out,
	output reg [63:0] VGA_buffer,
	output [2:0] counter_out, 
	output WR_ext, 
	output OE_core, 
	output OE_ext, 
	output reg activ_8_16, 
	output led_HALT
);

	reg [15:0] C_reg;
	reg [15:0] B_reg;
	reg [15:0] instr_reg;
	reg [15:0] immediate_reg;
	reg [15:0] PC_low;
	reg [15:0] PC_high;
	reg [15:0] PC_high_temp;
	reg [15:0] addr_A_high;
	reg [15:0] addr_A_low;
	reg [15:0] addr_B_high;
	reg [15:0] addr_B_low;
	reg [15:0] config_reg;
	reg [15:0] loop_counter;
	reg [1:0] VGA_micro_address;


	reg [2:0] seq_counter; 
	reg [1:0] sel_addr;
	reg cond_reg;
	reg cond_mux_out;
	reg cond_loop_counter;
	reg allow_copy;
	reg is_not_null;
	reg activ_immediate;
	reg [15:0] status_reg; 
	reg WR_cache;
	reg OE_core_pre;
	reg OE_core_post;
	reg OE_ext_pre;
	reg OE_ext_post;
	reg WR_ext_pre;
	reg WR_ext_post;
	reg reset_mem;
	
	//false registers
	reg [15:0] X_to_CPU; 
	reg [15:0] C_input; 
	reg [15:0] ALU_input; 
	reg [11:0] micro_addr_decode;
	reg [1:0] addr_VGA_low;
	reg [15:0] ext_to_CPU_8_16;
	reg addr_buff_0;
	
	reg activ_B;
	reg activ_C;
	reg activ_X;
	
	wire [2:0] micro_instr;
	wire instr_8_16;
	wire [3:0] ALU_op;
	wire [7:0] micro_addr;
	wire [15:0] ALU_out;
	wire [15:0] cache_to_CPU;
	wire cond_from_ALU;
	wire activ_cond_ALU;
	wire [23:0] PC_tot;
	wire [23:0] addr_A_tot;
	wire [23:0] addr_B_tot;
	wire [23:0] addr_VGA_tot;
	
	// end declarations
	
	//******************************
	// ****** CORE COMPONENTS ******
	core_cache core_cache_01(micro_addr[6:0],	clk_slow, X_from_CPU, WR_cache, cache_to_CPU);//clk_fast
	ALU_A2Z ALU_1(X_to_CPU, ALU_input, ALU_op, ALU_out, cond_from_ALU, activ_cond_ALU);
	assign led_HALT=!config_reg[0];
	assign OE_core = OE_core_post;
	assign OE_ext = OE_ext_pre;
	assign WR_ext = WR_ext_post && WR_ext_pre;
	
	assign addr_VGA_tot = {4'd4, addr_VGA_high, VGA_micro_address, 1'b0}; 
	assign counter_out = seq_counter; 
	
	addr_adder addr_adder_A(addr_A_low, addr_A_high, addr_A_tot);
	addr_adder addr_adder_B(addr_B_low, addr_B_high, addr_B_tot);
	assign PC_tot = {PC_high[15:9], PC_low,1'b0}; 
		
	// *****************************
	// **** INSTRUCTION DECODER ****
	assign micro_instr = instr_reg[15:13];
	assign instr_8_16 = instr_reg[12];
	assign ALU_op = instr_reg[11:8];
	assign micro_addr = instr_reg[7:0];
	
	always @(*) begin
		if ((micro_instr==3'b100)||(micro_instr==3'b110)||(micro_instr==3'b011))
			activ_immediate <= 1;
		else 
			activ_immediate <= 0;
	end
	
	always @(*) begin // micro address decoder (to_X only)
		if (micro_addr[7]==1'b1)
			micro_addr_decode <= 12'b1; // cache
		else begin 
			case (micro_addr[6:3])
				4'b1111 : micro_addr_decode <= 12'b0000_0000_0010; // B
				4'b0000 : micro_addr_decode <= 12'b0000_0000_0100; // ext (A ou B)
				4'b0001 : micro_addr_decode <= 12'b0000_0000_0100;	// ext (A ou B)
				4'b0101 : micro_addr_decode <= 12'b0000_0000_1000; // addr_A_high
				4'b0100 : micro_addr_decode <= 12'b0000_0001_0000; // addr_A_low
				4'b0111 : micro_addr_decode <= 12'b0000_0010_0000; // addr_B_high
				4'b0110 : micro_addr_decode <= 12'b0000_0100_0000; // addr_B_low
				4'b1001 : micro_addr_decode <= 12'b0000_1000_0000; // PC_high
				4'b1000 : micro_addr_decode <= 12'b0001_0000_0000; // PC_low
				4'b1100 : micro_addr_decode <= 12'b0010_0000_0000; // CONFIG
				4'b1110 : micro_addr_decode <= 12'b0100_0000_0000; // COUNTER
				default : micro_addr_decode <= 12'b0000_0000_0000;
			endcase
		end 
	end
	
	always @(*) begin // activ_B construction
		if (micro_instr==3'b001)
			activ_B <= 1;
		else 
			activ_B <= 0;
	end
	
	always @(*) begin	// activ_C construction 
		if ((micro_instr==3'b000)||(micro_instr==3'b010)||(micro_instr==3'b011))
			activ_C <= 1;
		else
			activ_C <= 0;
	end
	
	always @(*) begin
		if (loop_counter == 16'd1)
			cond_loop_counter <= 1'b1; //warning, inverted cond
		else 
			cond_loop_counter <= 1'b0; 
	end
	
	always @(*) begin		// MUX COND
		if (config_reg[1] == 1'b1)
			cond_mux_out <= cond_loop_counter;
		else 
			cond_mux_out <= cond_reg;
	end
	
	
	always @(*) begin
		if (C_reg == 16'd0)
			is_not_null <= 1'b0;
		else 
			is_not_null <= 1'b1;
	end
	
	always @(*) begin
		if (config_reg[6]==1'b1)
			allow_copy <= is_not_null;
		else 
			allow_copy <= 1'b1;
	end
	
	always @(*) begin // construction activ_X
		if ((micro_instr[2:1]==2'b10)||((micro_instr[2:1]==2'b11)&&(cond_mux_out==0))) //inverted cond
			activ_X <= 1;
		else 
			activ_X <= 0;
	end
	
	always @(*) begin
		case (seq_counter)
			3'd6 : activ_8_16 <= instr_8_16;
			3'd7 : activ_8_16 <= instr_8_16;
			default : activ_8_16 <= 1'b1;
		endcase
	end
	
	// **********************************
	// **** 		CORE DATAPATH MUX		****
	always @(*) begin // MUX X_from_CPU
		if (micro_instr[0]==1'b1) 
			X_from_CPU <= C_reg;
		else 
			X_from_CPU <= immediate_reg;
	end
	
	always @(*) begin // MUX C_input
		if (micro_instr[1]==1'b1)
			C_input <= ALU_out;
		else
			C_input <= X_to_CPU;
	end
	
	always @(*) begin // MUX ALU_input
		if (micro_instr[0]==1'b1)
			ALU_input <= immediate_reg;
		else 
			ALU_input <= B_reg;
	end
	
	always @(*) begin // MUX Adresse
		if (sel_addr == 2'b01) //PC
			addr_out <= PC_tot;
		else if (sel_addr == 2'b10) //VGA
			addr_out <= addr_VGA_tot;
		else begin
			if (micro_addr[7:3]==5'b00001) //Ext
				addr_out <= addr_B_tot;
			else if (micro_addr[7:3]==5'b00000)
				addr_out <= addr_A_tot;
			else 
				addr_out <= 24'hFF_FF_FF;
		end
	end
	
	always @(*) begin // MUX X_to_CPU
		if (micro_addr[7]==1)
			X_to_CPU <= cache_to_CPU;
		else begin 
			if (micro_addr[6:3]==4'b1101)
				X_to_CPU <= status_reg;
			else 
				X_to_CPU <= ext_to_CPU_8_16;
		end
	end
	
	always @(*) begin // 8 16 bits ext_to_CPU
		if (instr_8_16 == 1'b1) 
			ext_to_CPU_8_16 <= ext_to_CPU;
		else begin
			ext_to_CPU_8_16[15:8] <=8'd0;
			if (addr_buff_0==1'b0)
				ext_to_CPU_8_16[7:0] <= ext_to_CPU[7:0];
			else 
				ext_to_CPU_8_16[7:0] <= ext_to_CPU[15:8];
		end
	end
	
	// ************************
	// **** CORE SEQUENCER ****
	always @(posedge clk_fast) begin
		OE_ext_post <= OE_ext_pre;
		OE_core_post <= OE_core_pre;
		WR_ext_post <= WR_ext_pre;
	end
	
	always @(posedge clk_slow) begin
		addr_buff_0 <= addr_out[0];
	end
	
	always @(posedge clk_slow) begin
		seq_counter <= seq_counter+4'd1;
		reset_mem <= reset_in;
		
		// real sequencer
		if (seq_counter==3'd0) begin
			sel_addr <= 2'b10;
			VGA_micro_address <=2'b01; //2'b00
		end
		
		if (seq_counter==3'd0) begin //T0
			OE_ext_pre <= 1'b1;
			if (config_reg[0]==1'b0) begin //final store
				if ((micro_instr[2:1] == 2'b01)&&(activ_cond_ALU == 1'b1))
					cond_reg<=cond_from_ALU;
				if (activ_C==1)
					C_reg <= C_input;
				if (activ_B==1)
					B_reg <= X_to_CPU;
			end
		end
		
		if (seq_counter == 3'd1) begin 			//T1 Load VGA 0&1
			VGA_micro_address <=2'b10; //2'b01
			VGA_buffer[15:0] <= ext_to_CPU;
			if (reset_mem==1'b0) begin //Async RESET
				PC_low <= 16'h0000;
				PC_high <= 16'h0000;
				config_reg <= 16'd0;
			end
			if (micro_addr_decode[8]==1'b1)//PC
				loop_counter <= loop_counter - 16'd1;
			if ((micro_addr[7:3]==5'b0_0000) && (config_reg[2]==1'b1)) //post increment A
				addr_A_low <= addr_A_low + 16'd1;
			if ((micro_addr[7:3]==5'b0_0001) && (config_reg[4]==1'b1)) //post increment B
				addr_B_low <= addr_B_low + 16'd1;
		end
			
		if (seq_counter == 3'd2) begin		//T2 Load VGA 2&3
			VGA_micro_address <=2'b11; //2'b10
			VGA_buffer[31:16] <= ext_to_CPU;
			if  ((micro_addr[7:3]==5'b0_0000) && (config_reg[3]==1'b1)) //post increment A
				addr_A_low <= addr_A_low + 16'd1;
			if ((micro_addr[7:3]==5'b0_0001) && (config_reg[5]==1'b1)) //post increment B
				addr_B_low <= addr_B_low + 16'd1;
		end
			
		if (seq_counter == 3'd3) begin 			//T3 Load VGA 4&5
			sel_addr <= 2'b01; // sel_PC
			VGA_buffer[47:32] <= ext_to_CPU;
			VGA_micro_address <=2'b00;
		end
		
		if (seq_counter == 3'd4)  //T4
			VGA_buffer[63:48] <= ext_to_CPU;

		if ((seq_counter == 3'd4)&&((config_reg[0]==1'b0))) //T4 
			PC_low <= PC_low+16'd1;
		
		if ((seq_counter==3'd5)&&(config_reg[0]==1'b0)) begin	// T5 load instruction
			instr_reg <= ext_to_CPU;
			sel_addr <= 2'b00; //sel_ext
		end
			
		if ((seq_counter==3'd6)&&(config_reg[0]==1'b0)) begin // T6 immediate_reg loading; begin compute
			if (activ_immediate==1) begin 
				immediate_reg <= ext_to_CPU;
				PC_low <= PC_low+16'd1; 
			end
			if (activ_X==1) begin 
				if ((micro_addr_decode[2]==1'b1) && (allow_copy==1'b1)) begin
					OE_ext_pre <= 1'b0;
					OE_core_pre <= 1'b1;
					WR_ext_pre <= 1'b1;
				end 
				if (micro_addr_decode[0]==1'b1)
					WR_cache <=1;
			end
		end
		
		if (seq_counter==3'd7)
			sel_addr <= 2'b10; //sel_VGA
			
		if ((seq_counter==3'd7)&&(config_reg[0]==1'b0)) begin //final store
			OE_core_pre <= 1'b0;
			if (activ_X==1) begin 	// TO X
				if (micro_addr_decode[0]==1'b1)
					WR_cache <= 1'b0;
				if (micro_addr_decode[1]==1'b1)
					B_reg <= X_from_CPU;
				if (micro_addr_decode[2]==1'b1)
					WR_ext_pre <= 1'b0;
				if (micro_addr_decode[3]==1'b1)
					addr_A_high <= X_from_CPU;
				if (micro_addr_decode[4]==1'b1)
					addr_A_low <= X_from_CPU;
				if (micro_addr_decode[5]==1'b1)
					addr_B_high <= X_from_CPU;
				if (micro_addr_decode[6]==1'b1)
					addr_B_low <= X_from_CPU;
				if (micro_addr_decode[7]==1'b1)
					PC_high_temp <= X_from_CPU;
				if (micro_addr_decode[8]==1'b1) begin 
					PC_low <= X_from_CPU;
					PC_high <= PC_high_temp;
				end
				if (micro_addr_decode[9]==1'b1)
					config_reg <= X_from_CPU;
				if (micro_addr_decode[10]==1'b1)
					loop_counter <= X_from_CPU;
			end
		end	
	end
	
endmodule