module top_A2Z (
	input clk_in_l,
	input reset_in_l,
	input PS2_clk_pin,
	input PS2_data_pin,
	input SPI_in_pin,
	input [1:0] GP_IN,
	output led_HALT,
	output H_sync, 
	output V_sync, 
	output [3:0] red, 
	output [3:0] blue, 
	output [3:0] green,
	output [18:0] address_out,
	output reg [4:1] CS_out,
	output OE_out,
	output WR_out,
	output [1:0] GP_OUT,
	output SPI_clk,
	output SPI_out_pin,
	inout tri [15:0] data_out
);
    
	wire [23:0] address;
	wire [15:0] ext_from_CPU;
	reg [15:0] ext_to_CPU;
	reg [23:0] addr_buff;
	reg [1:0] GP_IN_buff;
	wire [15:0] from_boot;
	wire [16:0] addr_VGA;
	reg [1:0] GP_OUT_buf;
	wire activ_write;
	wire OE_core;
	wire OE_ext;
	wire PLL_1_locked;
	wire clock_slow;
	wire clock_fast;
	wire clock_fast_useless;
	wire reset_core;
	wire [63:0] VGA_buffer; 
	wire [2:0] micro_counter;
	wire PS2_read_in_progress;
	wire [7:0] PS2_FIFO_in; 
	wire [7:0] PS2_FIFO_out; 
	wire [7:0] SPI_data_RX;
	wire PS2_data_ready;
	wire PS2_FIFO_WR;
	wire PS2_FIFO_RD;
	wire instr_8_16;
	wire PS2_FIFO_empty;
	wire [15:0] ext_from_CPU_8_16; 
	wire displayed_frame;
	
	reg activ_write_boot, activ_write_txt;
	reg activ_write_SPI;
	reg select_ext;
	reg select_SRAM_12;
	reg select_SRAM_34;
	reg select_SRAM_1;
	reg select_SRAM_2;
	reg select_SRAM_3;
	reg select_SRAM_4;
	reg select_boot;
	reg select_text;
	reg select_boot_post;
	reg select_text_post;
	reg select_frame;
	reg select_displayed;
	reg instr_8_16_post;
	reg select_PS2_data;
	reg select_PS2_flag;
	reg select_GPIO_IN;
	reg select_GPIO_OUT;
	reg select_SPI_write;
	reg select_SPI_write_post;
	reg [6:0] local_select;
	reg PS2_read_in_progress_post;
	reg requested_frame;
	reg reset_buff;
	
	main_pll main_PLL_low (clk_in_l, clock_slow, clock_fast, PLL_1_locked);
	
	assign reset_core = (PLL_1_locked && reset_in_l);
	
	core_A2Z core_01 (
		clock_slow, clock_fast, reset_core, addr_VGA, ext_to_CPU, 
		ext_from_CPU, address, VGA_buffer, micro_counter, 
		activ_write, OE_core, OE_ext, instr_8_16, led_HALT);
	
	always @(posedge clock_slow) begin
		addr_buff <= address;
		CS_out <= {!select_SRAM_4,!select_SRAM_3,!select_SRAM_2,!select_SRAM_1};
		local_select <= {select_displayed, select_GPIO_IN, select_SPI_write, select_PS2_flag, select_PS2_data, select_ext, select_boot};
		select_boot_post <= select_boot;
		select_text_post <= select_text;
		instr_8_16_post <= instr_8_16;
		PS2_read_in_progress_post <= PS2_read_in_progress;
		select_SPI_write_post <= select_SPI_write;
		GP_IN_buff <= GP_IN;
		reset_buff <= reset_core;
	end
	
	Boot_RAM boot_RAM01(address[10:1], clock_slow, ext_from_CPU, activ_write_boot, from_boot);//clk_fast
	
	essai_VGA essai_VGA1(clock_slow, 
		reset_in_l, ext_from_CPU[7:0], address[11:0], activ_write_txt, 
		micro_counter, VGA_buffer, 
		requested_frame, displayed_frame, addr_VGA, 
		H_sync, V_sync, red, blue, green);
	
	assign PS2_read_in_progress = (select_PS2_data && OE_ext);
	
	assign PS2_FIFO_RD = ( PS2_read_in_progress && (!PS2_read_in_progress_post) );
	
	PS2_A2Z PS2_A2Z_1 (clock_slow, PS2_clk_pin, PS2_data_pin, 
		PS2_FIFO_in, PS2_FIFO_WR);

	
	PS2_FIFO PS2_FIFO_1(clock_slow, PS2_FIFO_in, PS2_FIFO_RD, PS2_FIFO_WR, PS2_FIFO_empty, PS2_FIFO_out);

	assign PS2_data_ready = !PS2_FIFO_empty; 
	
	A2Z_SPI A2Z_SPI_1(clock_slow, SPI_in_pin, ext_from_CPU[7:0], activ_write_SPI, 
	SPI_clk, SPI_out_pin, SPI_data_RX); 
	
	assign address_out=addr_buff[19:1];
	assign OE_out=!OE_ext;
	assign WR_out=!activ_write;

	always @(posedge clock_slow) begin
		if ( (select_GPIO_OUT==1'b1) && (activ_write==1'b1) )
			GP_OUT_buf <= ext_from_CPU[1:0];
	end
	
	always @(posedge clock_slow) begin
		if (reset_buff==1'b0)
			requested_frame <= 1'b0;
		else if ( (select_frame == 1'b1) && (activ_write==1'b1) )
			requested_frame <= ext_from_CPU[0];
	end
	
	assign GP_OUT = ~GP_OUT_buf;
	
	// address decoders
	always @(*) begin 
		if (address[23:22]==2'b01)	select_ext <= 1'b1; 
		else select_ext <= 1'b0;
	end
	
	always @(*) begin 
		if (address[23:20]==4'd4)	select_SRAM_12 <= 1'b1;
		else select_SRAM_12 <= 1'b0;
	end
	
	always @(*) begin 
		if (address[23:20]==4'd5)	select_SRAM_34 <= 1'b1;
		else select_SRAM_34 <= 1'b0;
	end
	
	always @(*) begin 
		if (address[23:12]==12'h001)	select_text <= 1'b1;
		else select_text <= 1'b0;
	end
	
	always @(*) begin 
		if (address[23:11]==13'h0000)	select_boot <= 1'b1;
		else select_boot <= 1'b0;
	end
	
	always @(*) begin 
		if (address[23:12]==12'h002)	select_PS2_data <= 1'b1;
		else select_PS2_data <= 1'b0;
	end
	
	always @(*) begin 
		if (address[23:12]==12'h003)	select_PS2_flag <= 1'b1;
		else select_PS2_flag <= 1'b0;
	end
	
	always @(*) begin 
		if (address[23:12]==12'h005)	select_SPI_write <= 1'b1;
		else select_SPI_write <= 1'b0;
	end
	
	always @(*) begin 
		if (address[23:12]==12'h006)	select_GPIO_IN <= 1'b1;
		else select_GPIO_IN <= 1'b0;
	end

	always @(*) begin 
		if (address[23:12]==12'h007)	select_GPIO_OUT <= 1'b1;
		else select_GPIO_OUT <= 1'b0;
	end

	always @(*) begin 
		if (address[23:12]==12'h008)	select_displayed <= 1'b1;
		else select_displayed<= 1'b0;
	end
	
		always @(*) begin 
		if (address[23:12]==12'h009)	select_frame <= 1'b1;
		else select_frame <= 1'b0;
	end

	always @(*) begin
		if ((select_text_post==1'b1)&&(activ_write==1'b1))	activ_write_txt <= 1'b1;
		else activ_write_txt <= 1'b0;
	end 
	
	always @(*) begin
		if ((select_SPI_write_post==1'b1)&&(activ_write==1'b1))	activ_write_SPI <= 1'b1;
		else activ_write_SPI <= 1'b0;
	end 
	
	always @(*) begin
		if ((select_boot_post==1'b1)&&(activ_write==1'b1))	activ_write_boot <= 1'b1;
		else activ_write_boot <= 1'b0;
	end 
	
	// tri state data port
	assign ext_from_CPU_8_16[7:0] = ext_from_CPU[7:0];
	
	assign ext_from_CPU_8_16[15:8] = instr_8_16_post ? ext_from_CPU[15:8] : ext_from_CPU[7:0];
	
	always @(*) begin 
		if ((select_SRAM_12==1'b1)&&((instr_8_16==1'b1)||((instr_8_16==1'b0)&&(address[0]==1'b0))))
			select_SRAM_1=1'b1; 
		else 
			select_SRAM_1=1'b0;
	end
	
	always @(*) begin 
		if ((select_SRAM_12==1'b1)&&((instr_8_16==1'b1)||((instr_8_16==1'b0)&&(address[0]==1'b1))))
			select_SRAM_2=1'b1; 
		else 
			select_SRAM_2=1'b0;
	end
	
	always @(*) begin 
		if ((select_SRAM_34==1'b1)&&((instr_8_16==1'b1)||((instr_8_16==1'b0)&&(address[0]==1'b0))))
			select_SRAM_3=1'b1; 
		else 
			select_SRAM_3=1'b0;
	end
	
	always @(*) begin 
		if ((select_SRAM_34==1'b1)&&((instr_8_16==1'b1)||((instr_8_16==1'b0)&&(address[0]==1'b1))))
			select_SRAM_4=1'b1; 
		else 
			select_SRAM_4=1'b0;
	end
	
	assign data_out = OE_core ? ext_from_CPU_8_16 : 16'hZZZZ; 
	
	// mux ext_to_CPU
	always @(*) begin
		case (local_select)
			7'b000_0001 : 		ext_to_CPU <= from_boot; 
			7'b000_0010 : 		ext_to_CPU <= data_out;
			7'b000_0100 : 		ext_to_CPU <= {8'd0,PS2_FIFO_out};
			7'b000_1000 : 		ext_to_CPU <= {15'd0,PS2_data_ready};
			7'b001_0000 : 		ext_to_CPU <= {8'd0,SPI_data_RX};
			7'b010_0000 : 		ext_to_CPU <= {14'd0,GP_IN_buff};
			7'b100_0000 : 		ext_to_CPU <= {15'd0,displayed_frame};
			default : 	ext_to_CPU <= 16'h0000;
		endcase
	end
	
endmodule
	