#include "stdafx.h"
#include "stdio.h"
#include "string.h"
#include "ctype.h"
#include <stdlib.h>
#include <SFML/Graphics.hpp>
#pragma warning(disable: 4996)

unsigned char VRAM[0x200000];
unsigned char Text_RAM[4096];
unsigned short V_CACHE[128];
int sim_time;

//all registers
unsigned short PC_H;
unsigned short PC_H_temp;
unsigned short PC_L;
unsigned short C_reg;
unsigned short B_reg;
unsigned short addr_A_H;
unsigned short addr_A_L;
unsigned short addr_B_H;
unsigned short addr_B_L;
unsigned short config_reg;
unsigned short counter_reg;
unsigned short frame_select=0;
unsigned short frame_displayed = 0;
unsigned char cond_reg;

unsigned char FIFO_content[256];
unsigned char FIFO_index_read = 0;
unsigned char FIFO_index_write = 0;
int FIFO_filling = 0;

//Flash emulation
unsigned char FF_buffer[65536];
unsigned char SPI_MISO_buf;
unsigned short SPI_CS;
unsigned short SPI_previous_CS;
unsigned long int current_flash_pos;
unsigned char SPI_MOSI_buf[1000];
unsigned long int SPI_bytes_count;
int SPI_instruction;
int Flash_write_enable;
FILE* FLASH_file_point;

sf::RenderWindow window(sf::VideoMode(640, 496), "A2Z Emulator2");
sf::Font font;
sf::Text text_SF;
sf::Texture texture;
sf::Uint8 pixels[640*480*4];

sf::Sprite sprite;

void FIFO_write(unsigned char content) {
	FIFO_content[FIFO_index_write] = content;
	FIFO_filling++;
	FIFO_index_write++;
}

int FIFO_read() {
	unsigned char content;
	content = FIFO_content[FIFO_index_read];
	FIFO_filling--;
	FIFO_index_read++;
	return content;
}

void emulate_flash(int cs_or_write, unsigned char spi_write_val) {
	int size_write;
	// c_or_write : 1 if CS ; 2 if write
	
	if (cs_or_write == 1) { // CS manipulation 
		if ((SPI_previous_CS == 1) && (SPI_CS == 0)) { //falling edge
			SPI_bytes_count = 0; 
			SPI_instruction = 0;
		}
		if ((SPI_previous_CS == 0) && (SPI_CS == 1)) { //rising edge
			if ((Flash_write_enable == 1) && (SPI_instruction == 0x02)) { //Flash Page Program
				// warning, not 100% representative of the actual flash behavior!
				size_write = SPI_bytes_count - 4;
				if (size_write > 256) { size_write = 256; }
				fwrite((SPI_MOSI_buf + 4), 1, size_write, FLASH_file_point);
				Flash_write_enable = 0;
			}
			
		}
		SPI_previous_CS = SPI_CS;
	}

	if (cs_or_write == 2) { //MOSI Master to flash SPI
		if (SPI_bytes_count < 900) {
			SPI_MOSI_buf[SPI_bytes_count] = spi_write_val;
		}
		SPI_bytes_count++;
		if (SPI_bytes_count == 1) {//command decode
			SPI_instruction = spi_write_val;
			//Flash_write_enable = 0;
			if (SPI_instruction == 0x06) {//WRITE enable
				Flash_write_enable = 1;
			}
			if ((SPI_instruction == 0x03) || (SPI_instruction == 0x04)) {//read or write disable
				Flash_write_enable = 0;
			}
		}
		if (SPI_bytes_count == 4) {
			//end address received
			current_flash_pos = (0x10000 * SPI_MOSI_buf[1]) + (0x100 * SPI_MOSI_buf[2]) + SPI_MOSI_buf[3];
			if ((SPI_instruction == 0x20) && (Flash_write_enable==1)) {// subsector errase
				current_flash_pos = current_flash_pos & 0xFFF000;
				fseek (FLASH_file_point, current_flash_pos, SEEK_SET);
				fwrite(FF_buffer, 1, 0x1000, FLASH_file_point);
				Flash_write_enable = 0;
			}
			if ((SPI_instruction == 0xD8) && (Flash_write_enable == 1)) {// sector errase
				current_flash_pos = current_flash_pos & 0xFF0000;
				fseek(FLASH_file_point, current_flash_pos, SEEK_SET);
				fwrite(FF_buffer, 1, 0x10000, FLASH_file_point);
				Flash_write_enable = 0;
			}
			if ((SPI_instruction == 0x03) || (SPI_instruction == 0x02))  {// flash read or program: set file position
				fseek(FLASH_file_point, current_flash_pos, SEEK_SET);
			}
		}
		if ((SPI_bytes_count > 4) && (SPI_instruction == 0x03)){ //Flash READ
			fread(&SPI_MISO_buf, 1, 1, FLASH_file_point);
		}
		if ((SPI_bytes_count > 1) && (SPI_instruction == 0x05)){ //Status READ
			SPI_MISO_buf = 0; 
		}
	}
}

unsigned short int get_data(unsigned int loc_address, unsigned char activ_8_16) {
	unsigned short int output_data;
	if ((loc_address & 0xE00000) == 0x400000) {
		if (activ_8_16 == 0) {
			output_data = VRAM[(loc_address & 0x1FFFFF)];
		}
		else {
			output_data = VRAM[(loc_address & 0x1FFFFE)];
			output_data = output_data + 256 * VRAM[(loc_address & 0x1FFFFE) + 1];
		}

	} else if ((loc_address & 0xFFF000) == 0x002000) {
		output_data = FIFO_read();
	} 
	else if ((loc_address & 0xFFF000) == 0x003000) {
		if (FIFO_filling == 0) { output_data = 0; }
		else { output_data = 1; }
	} 
	else if ((loc_address & 0xFFF000) == 0x005000) {
		output_data = SPI_MISO_buf;
	}
	else if ((loc_address & 0xFFF000) == 0x008000) {
		output_data = frame_displayed;
	} 
	else {
		output_data = 0;
	}
	return output_data;
}

void set_data(unsigned int loc_address, unsigned char activ_8_16, unsigned short data_in) {
	if ((loc_address & 0xE00000) == 0x400000) {
		if (activ_8_16 == 0) {
			VRAM[(loc_address & 0x1FFFFF)] = data_in;
		}
		else {
			VRAM[(loc_address & 0x1FFFFE)] = data_in % 256;
			VRAM[(loc_address & 0x1FFFFE) + 1] = data_in / 256;;
		}

	}
	else if ((loc_address & 0xFFF000) == 0x001000) { // TEXT
		Text_RAM[loc_address & 0x000FFF] = data_in;
	}
	else if ((loc_address & 0xFFF000) == 0x005000) { // GP out (SPI CS)
		emulate_flash(2, (data_in & 0xFF) );
	}
	else if ((loc_address & 0xFFF000) == 0x007000) { // GP out (SPI CS)
		SPI_CS = ((~data_in) & 0x02) /2;
		emulate_flash(1, 0);
	}
	else if ((loc_address & 0xFFF000) == 0x009000) {
		frame_select = (data_in & 0x0001);
	}
}

unsigned short ALU_compute(unsigned short A_input, unsigned short B_input, unsigned char loc_op) {
	unsigned short ALU_output;
	if (loc_op == 0) { ALU_output = A_input - B_input; }
	if (loc_op == 1) { ALU_output = A_input + B_input; }
	if (loc_op == 3) { ALU_output = (A_input * B_input) / 0x10000; }
	if (loc_op == 4) { ALU_output = (A_input * B_input) % 0x10000; }
	if (loc_op == 5) { ALU_output = (A_input / B_input); }
	if (loc_op == 6) { ALU_output = (A_input % B_input); }
	if (loc_op == 7) { ALU_output = ~A_input; cond_reg = ALU_output & 1; }
	if (loc_op == 8) {
		if (A_input == B_input) { ALU_output = 1; } else { ALU_output = 0; }
		cond_reg = ALU_output & 1; 
	}
	if (loc_op == 9) {
		if (A_input > B_input) { ALU_output = 1; } else { ALU_output = 0; }
		cond_reg = ALU_output & 1;
	}
	if (loc_op == 0xA) {
		if (A_input >= B_input) { ALU_output = 1; } else { ALU_output = 0; }
		cond_reg = ALU_output & 1;
	}
	if (loc_op == 0xB) {
		if (A_input < B_input) { ALU_output = 1; } else { ALU_output = 0; }
		cond_reg = ALU_output & 1;
	}
	if (loc_op == 0xC) {
		if (A_input <= B_input) { ALU_output = 1; } else { ALU_output = 0; }
		cond_reg = ALU_output & 1;
	}
	if (loc_op == 0xD) {
		if (A_input != B_input) { ALU_output = 1; } else { ALU_output = 0; }
		cond_reg = ALU_output & 1;
	}
	if (loc_op == 0xE) {
		ALU_output = A_input & B_input;
		cond_reg = ALU_output & 1;
	}
	if (loc_op == 0xF) {
		ALU_output = A_input | B_input;
		cond_reg = ALU_output & 1;
	}
	return ALU_output;
}

void execute_single(void) {
	unsigned short instruction_reg;
	unsigned short immediate_reg;
	unsigned short X_temp;
	unsigned char micro_address;
	unsigned char ALU_OP;
	unsigned char micro_instruction;
	unsigned char cond_8_16;
	unsigned char direction;
	unsigned char activ_write;
	unsigned char cond_mux;
	
	instruction_reg = get_data( (PC_H & 0xFE00) * 0x100 + PC_L * 2, 1);
	PC_L++;
	
	micro_address = instruction_reg % 256;
	ALU_OP = (instruction_reg & 0x0F00) / 256; 
	cond_8_16 = (instruction_reg & 0x1000) / 0x1000;
	micro_instruction = (instruction_reg & 0xE000) / 0x2000;
	direction = (instruction_reg & 0x8000) / 0x8000;

	if ((micro_instruction == 4) || (micro_instruction == 6) || (micro_instruction == 3)) {//immediate
		immediate_reg = get_data((PC_H & 0xFE00) * 0x100 + PC_L * 2, 1);
		PC_L++;
	}
	if (direction == 1) { // to X
		activ_write = 0;
		//cond mux!
		
		if ((config_reg & 2) == 2) {
			if (counter_reg == 1) { cond_mux = 1;  }
			else { cond_mux = 0; }
		} else {
			cond_mux = cond_reg;
		}
		if (micro_instruction == 4) { X_temp = immediate_reg; activ_write = 1; }
		if (micro_instruction == 5) { X_temp = C_reg; activ_write = 1; }
		if ((micro_instruction == 6) && (cond_mux == 0)) { X_temp = immediate_reg; activ_write = 1; }
		if ((micro_instruction == 7) && (cond_mux == 0)) { X_temp = C_reg; activ_write = 1; }

		if (activ_write == 1) {
			if (micro_address == 0x78) { B_reg = X_temp; }
			if (micro_address == 0x00) {
				if (((config_reg & 0x40) == 0x00) || (((config_reg & 0x40) == 0x40) && (C_reg != 0))) {
					set_data(addr_A_H * 256 + addr_A_L, cond_8_16, X_temp);
				}
				if ((config_reg & 4) == 4) { addr_A_L++; }
				if ((config_reg & 8) == 8) { addr_A_L++; }
			}
			if (micro_address == 0x08) { 
				if (((config_reg & 0x40) == 0x00) || (((config_reg & 0x40) == 0x40) && (C_reg != 0))) {
					set_data(addr_B_H * 256 + addr_B_L, cond_8_16, X_temp); 
				}
				if ((config_reg & 0x10) == 0x10) { addr_B_L++; }
				if ((config_reg & 0x20) == 0x20) { addr_B_L++; }
			}
			if (micro_address == 0x28) { addr_A_H = X_temp; }
			if (micro_address == 0x20) { addr_A_L = X_temp; }
			if (micro_address == 0x38) { addr_B_H = X_temp; }
			if (micro_address == 0x30) { addr_B_L = X_temp; }
			if (micro_address == 0x48) { PC_H_temp = X_temp; }
			if (micro_address == 0x40) { 
				PC_H = PC_H_temp; 
				PC_L = X_temp;
				counter_reg--;
			}
			if (micro_address == 0x60) { config_reg= X_temp; }
			if (micro_address == 0x70) { counter_reg = X_temp; }
			if ((micro_address & 0x80) == 0x80) {// CACHE
				V_CACHE[micro_address & 0x7F] = X_temp;
			}
		}
	}
	if (direction == 0) { // from X
		// X elaboration
		if (micro_address == 0x00) {
			X_temp = get_data(addr_A_H * 256 + addr_A_L, cond_8_16);
			if ((config_reg & 4) == 4) { addr_A_L++; }
			if ((config_reg & 8) == 8) { addr_A_L++; }
		}
		if (micro_address == 0x08) {
			X_temp = get_data(addr_B_H * 256 + addr_B_L, cond_8_16);
			if ((config_reg & 0x10) == 0x10) { addr_B_L++; }
			if ((config_reg & 0x20) == 0x20) { addr_B_L++; }
		}
		if ((micro_address & 0x80) == 0x80) {// CACHE
			X_temp = V_CACHE[micro_address & 0x7F];
		}

		if (micro_instruction == 0) { C_reg = X_temp; }
		if (micro_instruction == 1) { B_reg = X_temp; }
		if (micro_instruction == 2) { C_reg = ALU_compute(X_temp, B_reg, ALU_OP); }
		if (micro_instruction == 3) { C_reg = ALU_compute(X_temp, immediate_reg, ALU_OP); }
	}

}



unsigned char key_code_lut_1[200] = { 88,21,50,33,35,36,43,52,51,67,59,66,75,76,49,68,77,28,45,27,44,60,42,26,34,53,29,69,22,30,38,37,46,54,61,62,70,118,20,18,17,31,20,89,17,39,47,78,84,91,58,65,224,73,93,82,85,0,41,90,102,13,125,122,105,108,112,113,121,123,124,74,107,116,117,114,112,105,114,122,107,115,116,108,117,125,5,6,4,12,3,11,131,10,1,9,120,7,0,0,0,0};
unsigned char key_code_lut_2[200] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,0,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,0,0,0,1,1,1,1,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
void manage_key_event(int raw_key_code, char released) {
	if (key_code_lut_2[raw_key_code+1] == 1) {
		FIFO_write(0xE0); // extended
	}
	if (released == 1) {
		FIFO_write(0xF0); //release
	}
	FIFO_write(key_code_lut_1[raw_key_code + 1]);
}

char raw_text[10000];
char temp_text[500];

char charset_conv[256] = { 32,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,128,129,-23,131,132,-32,134,-25,136,137,-24,139,140,141,142,143,144,145,146,147,148,149,150,-7,152,153,154,155,156,157,158,159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255};
void display_txt(char* filename) {
	int i,j;
	int time_seconds;
	char current_char;
	strcpy(raw_text, "");
	for (i = 0; i <= 29; i++){//29
		for (j = 0; j <= 79; j++) {
			current_char = Text_RAM[j + 128 * i];
			if (current_char >= 0) { temp_text[j] = charset_conv[current_char]; }
			else { temp_text[j] = charset_conv[current_char + 256]; }
		}
		temp_text[80] = '\n';
		strcat(raw_text, temp_text);
	}
	time_seconds = (sim_time / 60) % 60;
	if (time_seconds < 10) {
		sprintf(temp_text, "SIMU TIME %d:0%d   %s\n", sim_time / 3600, time_seconds, filename);
	} else {
		sprintf(temp_text, "SIMU TIME %d:%d   %s\n", sim_time / 3600, time_seconds, filename);
	}
	strcat(raw_text, temp_text);
	text_SF.setString(raw_text);
	text_SF.setPosition(sf::Vector2f(0, -8)); 
}

sf::Uint8 palette_R[260] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240};
sf::Uint8 palette_G[260] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,48,64,80,32,48,64,80,32,48,64,80,32,48,64,80,64,96,128,160,64,96,128,160,64,96,128,160,64,96,128,160,96,144,192,240,96,144,192,240,96,144,192,240,96,144,192,240};
sf::Uint8 palette_B[260] = { 0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240,0,0,0,0,32,48,64,80,64,96,128,160,96,144,192,240};

void display_graph (void) {
	int x, y;
	unsigned int ram_address;
	unsigned int pixels_address;
	unsigned char raw_color;
	unsigned int offset_frame;
	// double buffer management
	if (frame_select == 1) { 
		offset_frame = 0x5A000; 
		frame_displayed = 1;
	}
	else { 
		offset_frame = 0; 
		frame_displayed = 0;
	}
	for (y = 0; y <= 479; y++) {
		ram_address = y * 0x300 + offset_frame;
		pixels_address = y * 2560; //640x4=2560
		for (x = 0; x <= 639; x++) {
			raw_color = VRAM[ram_address];
			pixels[pixels_address] = palette_R[raw_color];
			pixels[pixels_address + 1] = palette_G[raw_color];
			pixels[pixels_address + 2] = palette_B[raw_color];
			pixels[pixels_address + 3] = 0xFF;
			ram_address++;
			pixels_address = pixels_address + 4;//4
		}
	}
	texture.update(pixels);
	sprite.setTexture(texture);
}

unsigned char raw_input[3000000];
int read_input_file(char* file_name, int activate_goto) { // equivalent to Bootloader. Takes .bin or .bine or .bina files
	FILE* FID_in;
	int charcount_in;
	int end_read;
	int err_read;
	int i,j;
	int d_offset;
	int d_size;
	int is_not_exec;
	unsigned char chksum;
	
	is_not_exec = 1;
	FID_in = 0;
	FID_in = fopen(file_name, "rb");
	if (FID_in == 0) {
		printf("ERROR: input file '%s' not found\r\n", file_name);
		exit(EXIT_FAILURE);
	}
	charcount_in=fread(raw_input, 1, 2999000, FID_in);
	fclose(FID_in);
	
	end_read = 0;
	err_read = 0;
	i = 0;
	while (end_read == 0) {
		if (i >= charcount_in) {
			end_read = 1;
			printf("end of input file %d\r\n", i);
		}
		else if (raw_input[i] != 0x7B) { //first_char
			printf("error first char pos %d value %x\r\n", i, raw_input[i]);
			err_read = 1;
			end_read = 1;
		}
		else {//valid first_char
			i++;
			chksum = 0;
			if (raw_input[i] == 0xC3) { //download
				i++;
				d_offset = raw_input[i];
				i++;
				d_offset = d_offset + 256 * raw_input[i];
				i++;
				d_size = raw_input[i];
				i++;
				d_size = d_size + 256 * raw_input[i];
				i++;
				for (j = 0; j < d_size; j++) { //content read
					set_data(d_offset * 256 + j, 0, raw_input[i]);
					chksum = chksum + raw_input[i];
					i++;
				}
				if (raw_input[i] != 0x7B){
					printf("err read input file : expecting 0x7B at end %x\r\n", i);
					err_read = 1;
					end_read = 1;
				}
				else { i++; }
				if (raw_input[i] != chksum){
					printf("err bad chksum %x\r\n", i);
					err_read = 1;
					end_read = 1;
				}
				else { i++; }
			} 
			else if (raw_input[i] == 0xE6) {//Goto
				i++;
				d_offset = raw_input[i];
				chksum = raw_input[i];
				i++;
				d_offset = d_offset + 256 * raw_input[i];
				chksum = chksum + raw_input[i];
				i++;
				printf("goto %x\r\n", d_offset);
				if (raw_input[i] != 0x7B){
					printf("err read input file : expecting 0x7B at end %x\r\n", i);
					err_read = 1;
					end_read = 1;
				}
				else { i++; }
				if (raw_input[i] != chksum){
					printf("err bad chksum %x\r\n", i);
					err_read = 1;
					end_read = 1;
				}
				else { 
					i++; 
					if (activate_goto == 1) {
						PC_L = 0;
						PC_H = d_offset;
						PC_H_temp = d_offset;
						config_reg = 0;
					}
					end_read = 1;
					is_not_exec = 0;
				}

			}

		}

	}
	if (err_read == 1) {
		printf("there was an error reading the input file\r\n");
	}
	return is_not_exec;
}

void SRAM_dump(void) {
	FILE* FID_out;

	FID_out = fopen("SRAM_dump.dmp", "wb");
	if (FID_out == 0) {
		printf("ERROR: dump file is already opened by another application\r\n");
		exit(EXIT_FAILURE);
	}
	fwrite(VRAM, 1, 0x200000, FID_out);
	fclose(FID_out);
	printf("SRAM_dump.dmp successfully generated\r\n");
}

int get_new_input_file(void) {
	char filename[200];
	char* return_char;
	int isnotexec;
	printf("\r\nENTER NEW BIN FILE : \r\n");
	
	window.setVisible(false);
	fgets(filename, 200, stdin);
	return_char = strchr(filename, '\n');
	return_char[0] = 0;
	isnotexec = read_input_file(filename, 0);
	window.setVisible(true);
	return isnotexec;
}

void load_from_flash(void) {
	fseek(FLASH_file_point, 0x80000, SEEK_SET);
	fread(VRAM + 0xC0000, 1, 0x10000, FLASH_file_point);
	printf("OS loaded from flash.\r\n");
	PC_H = 0x4C00;
	PC_H_temp = 0x4C00;
	PC_L = 0;
	config_reg = 0;
}

int main(int argc, char* argv[])
{
	char filename_in[200];
	char filename_in_trunk[200];
	char filename_out[200];
	char window_title[200];
	char footer_text[200];
	
	int pause_exec;
	int window_active;
	int refresh_time;
	int i, j;
	
	sf::Clock clock;
	sf::Time elapsed_time;
	sf::Event event;
	
	pause_exec = 1;
	for (i = 0; i <= 0xFFFF; i++) {
		FF_buffer[i] = 0xFF;
	}
	for (i = 0; i < 128; i++) {
		V_CACHE[i] = 0;
	}

	for (i = 1; i < argc; i++) {
		strcpy(filename_in, argv[i]);
		pause_exec = read_input_file(filename_in, 1);
	}
	FLASH_file_point = fopen("flash_content.A2Zf", "r+b");
	if (FLASH_file_point == 0) {
		printf("ERROR: 'flash_content.A2Zf' is already opened by another application\r\n");
		exit(EXIT_FAILURE);
	}
	// load from flash
	if (argc == 1) {
		load_from_flash();
		pause_exec = 0;
	}

	texture.create(640, 480);

	if (strlen(filename_in) > 35) {
		strcpy(filename_in_trunk, "[..]");
		strcat(filename_in_trunk, filename_in+strlen(filename_in) - 35);

	}
	else {
		strcpy(filename_in_trunk, filename_in);
	}
	font.loadFromFile("clacon_8x16.ttf");
	window.setPosition(sf::Vector2i(370, 50));
	strcpy(window_title, "A2Z Emulator   ");
	strcat(window_title, filename_in_trunk);
	strcat(window_title, "   (F4:Pause F5:Resume F7:Reboot F9:Load new)");
	window.setTitle(window_title);

	j = 0;
	sim_time = 0;
	
	text_SF.setFont(font);
	text_SF.setCharacterSize(20);
	text_SF.setPosition(sf::Vector2f(0, -8)); //500 480
	clock.restart();

	while (window.isOpen())
	{
		j++;
		if (pause_exec == 0) {
			sim_time++; 
			for (i = 0; i <= 52083; i++) { // 16ms (60Hz) of emulation at 25MHz = 52083 instructions
				execute_single();
			}
		}

		window_active = 0;

		while (window.pollEvent(event))
		{
			window_active = 1;
			if (event.type == sf::Event::Closed)
				window.close();
			if (event.type == sf::Event::KeyPressed) {
				manage_key_event(event.key.code,0);
			}
			if (event.type == sf::Event::KeyReleased) {
				manage_key_event(event.key.code, 1);
				if (event.key.code == 88) { // F4 key
					pause_exec = 1;
				}
				if (event.key.code == 89) { // F5 key
					pause_exec = 0;
				}
				if (event.key.code == 91) { // F7 key
					if (argc > 1) {
						read_input_file(filename_in, 1);
					} else {
						load_from_flash();
					}
					pause_exec = 0;
					sim_time = 0;
				}
				if (event.key.code == 93) { // F9 key
					pause_exec = get_new_input_file();
				}
			}
		}

		strcpy(footer_text, filename_in_trunk);
		if (pause_exec == 1) { strcat(footer_text, "   PAUSE"); }
		if ((pause_exec == 0) && (sim_time < 60)) { strcat(footer_text, "   BOOT"); }
				
		display_txt(footer_text);
		display_graph();
		window.clear();
		window.draw(sprite);
		window.draw(text_SF);
		window.display();
				
		if ((window_active == 1) || (pause_exec == 0)) { refresh_time = 16; }
		else { refresh_time = 200; }

		elapsed_time = clock.getElapsedTime();
		while (elapsed_time.asMilliseconds() < refresh_time) {
			sf::sleep(sf::milliseconds(1));
			elapsed_time = clock.getElapsedTime();
		}
		clock.restart();
	}
	SRAM_dump();
	fclose(FLASH_file_point);
	return 0;
}

