// bin_generator.cpp: 

#include "stdafx.h"
#include "stdio.h"
#include "string.h"
#include "ctype.h"
#include <stdlib.h>

char mif_txt_raw[2000000];
void read_MIF_file (char* filename_tot, int size_req, unsigned char* data_read) {
	FILE* fid_in;
	int i;
	int pos_raw;
	int current_addr;
	
	char* current_line;
	unsigned int raw_addr;
	unsigned int raw_2bits;
	for (i = 0; i <= size_req; i++) { //fill with Zeros by default
		data_read[i] = 0;
	}
	fid_in = fopen(filename_tot, "r");
	if (fid_in == 0) {
		printf("ERROR: input file '%s' not found\r\n", filename_tot);
		exit(EXIT_FAILURE);
	}
	fread(mif_txt_raw, 1, 399000, fid_in);
	current_line = strtok(mif_txt_raw, "\n");
	current_addr = -1;
	while (current_line != 0) {
		if (current_line[0] == '\t') {
			sscanf(current_line, "\t%x  :   %x", &raw_addr, &raw_2bits);
			current_addr = raw_addr;
			data_read[2 * raw_addr] = raw_2bits % 256;
			data_read[(2 * raw_addr) + 1] = raw_2bits / 256;
		}
		
		current_line = strtok(NULL, "\n");
	}
	fclose(fid_in);
}

unsigned char data_raw[2000000]; //(1MB)1048576
int main(int argc, char* argv[]) {
	FILE* fid_in;
	FILE* fid_out;
	FILE* fid_config;
	char temp_char1[50];
	char *temp_char2;
	char filename[40];
	char filename_base[40];
	char filename_extension[40];
	char size_txt[40];
	char offset_txt[40];
	unsigned int offset_int;
	unsigned int local_offset;
	char char_temp[20];
	char outfile_name[40];
	char conf_file_name[40];
	int i;
	int size;
	int size_tot;
	int size_1;
	int size_2;
	int offset;
	int read_count;
	int in_file_type;
	int end_write;
	int size_remaining; 
	unsigned int input_pos;
	unsigned char chksum = 0;
	unsigned char loc_value[20];
	
	printf("\r\n** BIN GENERATOR **\r\n");
	if (argc != 2) {
		printf("ERROR need 1 arguments\r\n");
		exit(EXIT_FAILURE);
	}
	strcpy(filename, argv[1]);
	strcpy(temp_char1, argv[1]);
	temp_char2 = strtok(temp_char1, ".");
	strcpy(filename_base, temp_char2);
	temp_char2 = strtok(NULL, ".");
	strcpy(filename_extension, temp_char2);

	if (strcmp(filename_extension, "mif") == 0) {
		in_file_type = 2; // MIF
	}
	else {
		in_file_type = 1; // other than MIF
	}

	// *** CONFIG FILE READ ***
	strcpy(conf_file_name, filename_base);
	strcat(conf_file_name, ".conf");
	fid_config = fopen(conf_file_name, "r");
	if (fid_config == 0) {
		printf("ERROR: cannot open config file '%s'\r\n", conf_file_name);
		exit(EXIT_FAILURE);
	}
	fscanf(fid_config, "%x %d", &offset_int, &size_tot); // char_temp
	printf("    config file: '0x%x' size:'%d'\r\n", offset_int, size_tot);
	fclose(fid_config);
	
	if ((size_tot > 1048576) || (size_tot < 1)) { //524287
		printf("Invalid size\r\n");
		exit(EXIT_FAILURE);
	}
	
	// *** INPUT FILE READ ***
	
	if (in_file_type == 1) { //other than .mif read binary
		fid_in = fopen(filename, "rb");
		if (fid_in == 0) {
			printf("ERROR: input file '%s' not found\r\n", filename);
			exit(EXIT_FAILURE);
		}
		read_count = 0;
		read_count = fread(data_raw, 1, size_tot, fid_in);
		if (read_count != size_tot) {
			printf("ERROR: input file '%s' is smaller than expected\r\n", filename);
			printf("read %d bytes instead of %d bytes expected\r\n", read_count, size_tot);
			exit(EXIT_FAILURE);
		}
		fclose(fid_in);
	}
	else if (in_file_type == 2) { // MIF File
		read_MIF_file(filename, size_tot, data_raw);
	}

	// OUTPUT FILE opening or creation
	strcpy(outfile_name, filename_base);
	if (in_file_type == 2) { strcat(outfile_name, ".bine"); }
	else { strcat(outfile_name, ".bina"); }

	fid_out = fopen(outfile_name, "wb");
	if (fid_out == 0) {
		printf("ERROR: out file '%s' is already opened by another application\r\n", outfile_name);
		exit(EXIT_FAILURE);
	}
	
	// OUTPUT FILE content
	input_pos = 0;
	local_offset = offset_int;
	while (input_pos < size_tot) {
		loc_value[0] = 123; // begin char
		loc_value[1] = 195; // write char
		loc_value[2] = local_offset % 256;
		loc_value[3] = local_offset / 256;
		
		size_remaining = size_tot - input_pos;
		if (size_remaining > 256) { size_remaining = 256; } //256
		loc_value[4] = size_remaining % 256;
		loc_value[5] = size_remaining / 256;
		fwrite(loc_value, 1, 6, fid_out);// ANNEX

		chksum = 0;
		for (i = 1; i <= size_remaining; i++) {
			chksum = chksum + data_raw[input_pos];
			fwrite(data_raw + input_pos, 1, 1, fid_out);
			input_pos++;
		}
		loc_value[0] = 123;//123
		fwrite(loc_value, sizeof(char), 1, fid_out);// ANNEX
		loc_value[0] = chksum;
		fwrite(loc_value, sizeof(char), 1, fid_out);// ANNEX


		local_offset = local_offset + 1;
	}

	// OUTPUT FILE goto (bine from MIF only)
	if (in_file_type == 2) {
		loc_value[0] = 123; // begin char
		loc_value[1] = 230; // goto char
		loc_value[2] = offset_int % 256;
		loc_value[3] = offset_int / 256;
		loc_value[4] = 123;
		loc_value[5] = (loc_value[2] + loc_value[3]); // checksum
		fwrite(loc_value, 1, 6, fid_out); // ANNEX
	}
	
	fclose (fid_out);
	printf("Bin generation SUCCESSFUL\r\n\r\n\r\n");
	return 0;
}

